"""
This module contains functions to download external resources.

If you are using the resources for research, please cite the original
authors.
"""
import requests
import zipfile
import os

PROJECT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

RESOURCE_MAP = {
    # Hedges, https://github.com/words/hedges
    "hedges": {
        "link": "https://raw.githubusercontent.com/words/hedges/main/data.txt",
        "area": "Semantics",
        "subarea": "Hedges",
        "filename": "hedges.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Semantics",
                                 "Hedges", "hedges.txt"),
        "bibtex": rb"""@misc{hedges, author = {Titus Wormer}, title = {Hedges}, year = {2022}, publisher = {GitHub}, journal = {GitHub repository}, howpublished = {\url{github.com/words/hedges}}, }"""
    },
    # Warriner, A. B., Kuperman, V., & Brysbaert, M. (2013).
    # Norms of valence, arousal, and dominance for 13,915 English lemmas.
    # Behavior Research Methods, 45(4), 1191-1207.
    "vad_warriner": {
        "link": "https://static-content.springer.com/esm/"
                "art%3A10.3758%2Fs13428-012-0314-x/MediaObjects/"
                "13428_2012_314_MOESM1_ESM.zip",
        "area": "Emotion",
        "subarea": "VAD",
        "filename": "BRM-emot-submit.csv",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "VAD", "BRM-emot-submit.csv"),
        "bibtex": rb"""
        @article{vad_warriner,
        title={Norms of valence, arousal, and dominance for 13,915 English lemmas},
        author={Warriner, Amy Beth and Kuperman, Victor and Brysbaert, Marc},
        journal={Behavior research methods},
        volume={45},
        pages={1191--1207},
        year={2013},
        publisher={Springer}
        }
        """
    },
    # Mohammad, S. M. (2018).
    # Obtaining reliable human ratings of valence, arousal, and dominance for 
    # 20,000 English words.
    # In Proceedings of the 56th Annual Meeting of the Association for
    # Computational Linguistics (Volume 1: Long Papers) (pp. 174-184).
    "vad_nrc": {
        "link": "https://saifmohammad.com/WebDocs/Lexicons/"
                "NRC-VAD-Lexicon.zip",
        "area": "Emotion",
        "subarea": "VAD",
        "filename": "NRC-VAD-Lexicon/NRC-VAD-Lexicon.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "VAD", "NRC-VAD-Lexicon", 
                                 "NRC-VAD-Lexicon.txt"),
        "multilingual_filepath": os.path.join(PROJECT_PATH, "elfen_resources", 
                                              "Emotion", "VAD",
                                              "NRC-VAD-Lexicon",
                                              "NRC-VAD-Lexicon-"
                                              "ForVariousLanguages.txt"),
        "bibtex": rb"""
        @inproceedings{vad_nrc,
        title = "Obtaining Reliable Human Ratings of Valence, Arousal, and Dominance for 20,000 {E}nglish Words",
        author = "Mohammad, Saif",
        editor = "Gurevych, Iryna  and
        Miyao, Yusuke",
        booktitle = "Proceedings of the 56th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)",
        month = jul,
        year = "2018",
        address = "Melbourne, Australia",
        publisher = "Association for Computational Linguistics",
        url = "https://aclanthology.org/P18-1017",
        doi = "10.18653/v1/P18-1017",
        pages = "174--184",
        }
        """
    },
    # Mohammad, S. M. (2018)
    # Word affect intensity.
    # In Proceedings of the 56th Annual Meeting of the Association for
    # Computational Linguistics (Volume 1: Long Papers) (pp. 1609-1619).
    "intensity_nrc": {
        "link": "https://saifmohammad.com/WebDocs/Lexicons/"
                "NRC-Emotion-Intensity-Lexicon.zip",
        "area": "Emotion",
        "subarea": "Intensity",
        "filename": "NRC-Emotion-Intensity-Lexicon/"
                    "NRC-Emotion-Intensity-Lexicon-v1.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "Intensity", "NRC-Emotion-Intensity-Lexicon",
                                 "NRC-Emotion-Intensity-Lexicon-v1.txt"),
        "multilingual_filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                              "Emotion", "Intensity",
                                              "NRC-Emotion-Intensity-Lexicon",
                                              "NRC-Emotion-Intensity-"
                                              "Lexicon-ForVariousLanguages.txt"),
        "bibtex": rb"""
        @inproceedings{vad_nrc,
            title = "Word Affect Intensities",
            author = "Mohammad, Saif",
            editor = "Calzolari, Nicoletta  and
            Choukri, Khalid  and
            Cieri, Christopher  and
            Declerck, Thierry  and
            Goggi, Sara  and
            Hasida, Koiti  and
            Isahara, Hitoshi  and
            Maegaard, Bente  and
            Mariani, Joseph  and
            Mazo, H{\'e}l{\`e}ne  and
            Moreno, Asuncion  and
            Odijk, Jan  and
            Piperidis, Stelios  and
            Tokunaga, Takenobu",
            booktitle = "Proceedings of the Eleventh International Conference on Language Resources and Evaluation ({LREC} 2018)",
            month = may,
            year = "2018",
            address = "Miyazaki, Japan",
            publisher = "European Language Resources Association (ELRA)",
            url = "https://aclanthology.org/L18-1027",
        }
        """
    },
    # Brysbaert, M., Warriner, A. B., & Kuperman, V. (2014).
    # Concreteness ratings for 40 thousand generally known English word lemmas.
    # Behavior Research Methods, 46(3), 904-911.
    "concreteness_brysbaert": {
        "link": "https://static-content.springer.com/esm/"
                "art%3A10.3758%2Fs13428-013-0403-5/MediaObjects/"
                "13428_2013_403_MOESM1_ESM.xlsx",
        "area": "Psycholinguistics",
        "subarea": "Concreteness",
        "filename": "13428_2013_403_MOESM1_ESM.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Concreteness", "13428_2013_403_MOESM1_ESM.xlsx"),
        "bibtex": rb"""
        @article{concreteness_brysbaert,
        title={Concreteness ratings for 40 thousand generally known English word lemmas},
        author={Brysbaert, Marc and Warriner, Amy Beth and Kuperman, Victor},
        journal={Behavior research methods},
        volume={46},
        pages={904--911},
        year={2014},
        publisher={Springer}
        }
        """
    },
    # Brysbaert, M., Mandera, P., McCormick, S. F., & Keuleers, E. (2019).
    # Word prevalence norms for 62,000 English lemmas.
    # Behavior Research Methods, 51(2), 467-479.
    "prevalence_brysbaert": {
        "link": "https://static-content.springer.com/"
                "esm/art%3A10.3758%2Fs13428-018-1077-9/"
                "MediaObjects/13428_2018_1077_MOESM2_ESM.xlsx",
        "area": "Psycholinguistics",
        "subarea": "Prevalence",
        "filename": "13428_2018_1077_MOESM2_ESM.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Prevalence", "13428_2018_1077_MOESM2_ESM.xlsx"),
        "bibtex": rb"""
        @article{prevalence_brysbaert,
        title={Word prevalence norms for 62,000 English lemmas},
        author={Brysbaert, Marc and Mandera, Pawe{\l} and McCormick, Samantha F and Keuleers, Emmanuel},
        journal={Behavior research methods},
        volume={51},
        pages={467--479},
        year={2019},
        publisher={Springer}
        }
        """
    },
    # Kuperman, V., Stadthagen-Gonzalez, H., & Brysbaert, M. (2013).
    # Age-of-acquisition ratings for 30,000 English words.
    # Behavior Research Methods, 45(4), 1191-1207.
    "aoa_kuperman": {
        "link": "https://static-content.springer.com/esm/"
                "art%3A10.3758%2Fs13428-013-0348-8/"
                "MediaObjects/13428_2013_348_MOESM1_ESM.xlsx",
        "area": "Psycholinguistics",
        "subarea": "AgeOfAcquisition",
        "filename": "13428_2013_348_MOESM1_ESM.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "AgeOfAcquisition", "13428_2013_348_MOESM1_ESM.xlsx"),
        "bibtex": rb"""
        @article{aoa_kuperman,
        title={Age-of-acquisition ratings for 30,000 English words},
        author={Kuperman, Victor and Stadthagen-Gonzalez, Hans and Brysbaert, Marc},
        journal={Behavior research methods},
        volume={44},
        pages={978--990},
        year={2012},
        publisher={Springer}
        }
        """
    },
    # Baccianella, S., Esuli, A., & Sebastiani, F. (2010).
    # SentiWordNet 3.0: An enhanced lexical resource for sentiment
    # analysis and opinion mining.
    # In LREC (Vol. 10, pp. 2200-2204).
    "sentiwordnet": { # Currently not supported
        "link": "https://raw.githubusercontent.com/aesuli/SentiWordNet/"
                "master/data/SentiWordNet_3.0.0.txt",
        "area": "Emotion",
        "subarea": "Sentiment",
        "filename": "SentiWordNet_3.0.0.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "Sentiment", "SentiWordNet_3.0.0.txt"),
        "bibtex": rb""  # TODO
    },
    # Mohammad, S. M., & Turney, P. D. (2013).
    # Emotions evoked by common words and phrases:
    # Using Mechanical Turk to create an emotion lexicon.
    # In Proceedings of the NAACL HLT 2013 Workshop on
    # Computational Approaches to Analysis and Generation
    # of Emotion in Text (pp. 26-34).
    "sentiment_nrc": {
        "link": "https://saifmohammad.com/WebDocs/Lexicons/NRC-Emotion-Lexicon.zip",
        "area": "Emotion",
        "subarea": "Sentiment",
        "filename": "NRC-Emotion-Lexicon/NRC-Emotion-Lexicon-Wordlevel-v0.92.txt",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "Sentiment", "NRC-Emotion-Lexicon",
                                 "NRC-Emotion-Lexicon-Wordlevel-v0.92.txt"),
        "multilingual_filepath": os.path.join(PROJECT_PATH, "elfen_resources",
                                              "Emotion", "Sentiment",
                                              "NRC-Emotion-Lexicon",
                                              "NRC-Emotion-Lexicon-"
                                              "ForVariousLanguages.txt"),
        "bibtex": rb"""
        @article{sentiment_nrc,
        Author = {Mohammad, Saif M. and Turney, Peter D.},
        Journal = {Computational Intelligence},
        Number = {3},
        Pages = {436--465},
        Title = {Crowdsourcing a Word-Emotion Association Lexicon},
        Volume = {29},
        Year = {2013}
        }
        """
    },
    # Coso, B., Guasch, M., Buganovic, I., Ferre, P., & Hinojosa, J. A. (2022).
    # CROWD-5e: A croatian psycholinguistic database for affective norms for
    # five discrete emotions.
    # Behavior Research Methods, 55(1), 4018-4034.
    # TODO: Handle processing the data
    "affect_crowd5e": {
        "link": "https://figshare.com/ndownloader/files/36434421",
        "area": "Emotion",
        "subarea": "Affect",
        "filename": "CROWD-5e.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Emotion",
                                 "Affect", "CROWD-5e.xlsx"),
        "bibtex": rb""  # TODO
    },
    # Diveica, V, Pexman, P. M., & Binney, R. J. (2023).
    # Quantifying Social Semantics: An Inclusive Definition
    # of Socialness and Ratings for 8,388 English Words.
    # Behavior Research Methods, 55, 461-473.
    "socialness": {
        "link": "https://osf.io/download/29eyh/",
        "area": "Psycholinguistics",
        "subarea": "Socialness",
        "filename": "Socialness.csv",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Socialness", "Socialness.csv"),
        "bibtex": rb"""
        @article{socialness,
        title={Quantifying social semantics: An inclusive definition of socialness and ratings for 8388 English words},
        author={Diveica, Veronica and Pexman, Penny M and Binney, Richard J},
        journal={Behavior Research Methods},
        volume={55},
        number={2},
        pages={461--473},
        year={2023},
        publisher={Springer}
        }
        """
    },
    # Lynott, D., Connell, L., Brysbaert, M., Brand, J., & Carney, J. (2020).
    # The Lancaster Sensorimotor Norms: Multidimensional measures of
    # perceptual and action strength for 40,000 English words.
    # Behavior Research Methods, 52, 1271-1291.
    "sensorimotor_lancaster": {
        "link": "https://osf.io/download/pu85v/",
        "area": "Psycholinguistics",
        "subarea": "Sensorimotor",
        "filename": "LancasterSensorimotorNorms.xlsx",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Sensorimotor", "LancasterSensorimotorNorms.xlsx"),
        "bibtex": rb"""
        @article{sensorimotor_lancaster,
        title={The Lancaster Sensorimotor Norms: multidimensional measures of perceptual and action strength for 40,000 English words},
        author={Lynott, Dermot and Connell, Louise and Brysbaert, Marc and Brand, James and Carney, James},
        journal={Behavior research methods},
        volume={52},
        pages={1271--1291},
        year={2020},
        publisher={Springer}
        }
        """
    },
    # Winter, B., Lupyan, G., Perry, L. K., Dingemanse, M., & Perlman, M. (2021).
    # Iconicity ratings for 14,000 English words.
    # Behavior Research Methods, 56, 1640-1655.
    "iconicity_winter": {
        "link": "https://osf.io/download/ex37k/",
        "area": "Psycholinguistics",
        "subarea": "Iconicity",
        "filename": "WinterIconicityNorms.csv",
        "filepath": os.path.join(PROJECT_PATH, "elfen_resources", "Psycholinguistics",
                                 "Iconicity", "WinterIconicityNorms.csv"),
        "bibtex": rb"""
        @article{iconicity_winter,
        title={Iconicity ratings for 14,000+ English words},
        author={Winter, Bodo and Lupyan, Gary and Perry, Lynn K and Dingemanse, Mark and Perlman, Marcus},
        journal={Behavior research methods},
        volume={56},
        number={3},
        pages={1640--1655},
        year={2024},
        publisher={Springer}
        }
        """
    },
    # TODO: Find and add psycholinguistic norms for other languages?
}

LANGUAGES_NRC = {
    "af": "Afrikaans",
    "sq": "Albanian",
    "am": "Amharic",
    "ar": "Arabic",
    "hy": "Armenian",
    "az": "Azerbaijani",
    "eu": "Basque",
    "be": "Belarusian",
    "bn": "Bengali",
    "bs": "Bosnian",
    "bg": "Bulgarian",
    "ca": "Catalan",
    "ceb": "Cebuano",
    "ny": "Chichewa",
    "zh": "Chinese-Simplified",
    "zh-TW": "Chinese-Traditional",
    "co": "Corsican",
    "hr": "Croatian",
    "cs": "Czech",
    "da": "Danish",
    "nl": "Dutch",
    "eo": "Esperanto",
    "et": "Estonian",
    "fil": "Filipino",
    "fi": "Finnish",
    "fr": "French",
    "fy": "Frisian",
    "gl": "Galician",
    "ka": "Georgian",
    "de": "German",
    "el": "Greek",
    "gu": "Gujarati",
    "ht": "Haitian-Creole",
    "ha": "Hausa",
    "haw": "Hawaiian",
    "he": "Hebrew",
    "hi": "Hindi",
    "hmn": "Hmong",
    "hu": "Hungarian",
    "is": "Icelandic",
    "ig": "Igbo",
    "id": "Indonesian",
    "ga": "Irish",
    "it": "Italian",
    "ja": "Japanese",
    "jv": "Javanese",
    "kn": "Kannada",
    "kk": "Kazakh",
    "km": "Khmer",
    "rw": "Kinyarwanda",
    "ko": "Korean",
    "ku": "Kurdish-Kurmanji",
    "ky": "Kyrgyz",
    "lo": "Lao",
    "la": "Latin",
    "lv": "Latvian",
    "lt": "Lithuanian",
    "lb": "Luxembourgish",
    "mk": "Macedonian",
    "mg": "Malagasy",
    "ms": "Malay",
    "ml": "Malayalam",
    "mt": "Maltese",
    "mi": "Maori",
    "mr": "Marathi",
    "mn": "Mongolian",
    "my": "Myanmar-Burmese",
    "ne": "Nepali",
    "no": "Norwegian",
    "or": "Odia",
    "ps": "Pashto",
    "fa": "Persian",
    "pl": "Polish",
    "pt": "Portuguese",
    "pa": "Punjabi",
    "ro": "Romanian",
    "ru": "Russian",
    "sm": "Samoan",
    "gd": "Scots-Gaelic",
    "sr": "Serbian",
    "st": "Sesotho",
    "sn": "Shona",
    "sd": "Sindhi",
    "si": "Sinhala",
    "sk": "Slovak",
    "sl": "Slovenian",
    "so": "Somali",
    "es": "Spanish",
    "su": "Sundanese",
    "sw": "Swahili",
    "sv": "Swedish",
    "tg": "Tajik",
    "ta": "Tamil",
    "tt": "Tatar",
    "te": "Telugu",
    "th": "Thai",
    "tr": "Turkish",
    "tk": "Turkmen",
    "uk": "Ukrainian",
    "ur": "Urdu",
    "ug": "Uyghur",
    "uz": "Uzbek",
    "vi": "Vietnamese",
    "cy": "Welsh",
    "xh": "Xhosa",
    "yi": "Yiddish",
    "yo": "Yoruba",
    "zu": "Zulu"
}


def download_lexicon(link: str,
                     path: str,
                     filename: str = None
                     ) -> None:
    """
    Download a lexicon from a link and save it to a path.
    
    Args:
        link (str): Link to the lexicon.
        path (str): Path to save the lexicon.
        filename (str): Name of the file to save the lexicon.

    Returns:
        None
    """
    # Headers to avoid 406 response
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64;"
                      " rv:91.0) Gecko/20100101 Firefox/91.0",
        "Accept": "text/html,application/xhtml+xml,application/"
                  "xml;q=0.9,image/webp,*/*;q=0.8",
        "Accept-Language": "en-US,en;q=0.5",
        "Accept-Encoding": "gzip, deflate, br",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1",
        "Pragma": "no-cache",
        "Cache-Control": "no-cache"
    }
    response = requests.get(link, headers=headers)

    if filename is None:
        filename = link.split("/")[-1]

    if link.endswith(".zip"):
        with open("temp.zip", "wb") as f:
            f.write(response.content)
        with zipfile.ZipFile("temp.zip", "r") as zip_ref:
            zip_ref.extractall(path)
        os.remove("temp.zip")
    elif link.endswith(".xlsx"):
        filename = link.split("/")[-1]
        with open(os.path.join(path, filename), "wb") as f:
            f.write(response.content)
    elif link.endswith(".txt"):
        with open(os.path.join(path, filename), "wb") as f:
            f.write(response.content)
    else:
        with open(os.path.join(path, filename), "wb") as f:
            f.write(response.content)
    
def get_resource(feature: str) -> None:
    """
    Download a resource from the RESOURCE_MAP.

    Args:
        feature (str): Name of the feature to download.

    Returns:
        None
    """
    if feature not in RESOURCE_MAP:
        raise ValueError(f"Feature {feature} not found in RESOURCE_MAP.")

    # Making sure all the necessary directories exist
    os.makedirs(os.path.join(PROJECT_PATH, "elfen_resources",
                             RESOURCE_MAP[feature]["area"],
                             RESOURCE_MAP[feature]["subarea"]),
                             exist_ok=True)
    # Downloading the lexicon if it does not exist
    if not os.path.exists(RESOURCE_MAP[feature]["filepath"]):
        download_lexicon(RESOURCE_MAP[feature]["link"],
                         os.path.join(PROJECT_PATH, "elfen_resources",
                                      RESOURCE_MAP[feature]["area"],
                                      RESOURCE_MAP[feature]["subarea"]),
                                      RESOURCE_MAP[feature]["filename"])

def list_external_resources() -> None:
    """
    List all the external resources available in the RESOURCE_MAP.

    Args:
        None

    Returns:
        None
    """
    for feature in RESOURCE_MAP:
        print(f"Feature: {feature}")
        print("\n")

def get_bibtex() -> str:
    """
    Print the bibtex citation for all the resources in the RESOURCE_MAP,
    and the package itself.

    The citation keys for the lexicons are the names of the lexicons in
    the RESOURCE_MAP.
    The citation key for the package is "maurer-2024-elfen".

    Args:
        None

    Returns:
        bibxtex (str): Bibtex citation for all the resources.
    """

    bibtex = rb"""
    @misc{maurer-2024-elfen,
        author = {Maurer, Maximilian},
        title = {ELFEN - Efficient Linguistic Feature Extraction for Natural Language Datasets},
        year = {2024},
        publisher = {GitHub},
        journal = {GitHub repository},
        howpublished = {\url{https://github.com/mmmaurer/elfen}},
    }
    """
    for feature in RESOURCE_MAP:
        bibtex += RESOURCE_MAP[feature]["bibtex"]

    return bibtex.decode("utf-8")




"""
This module contains functions to preprocess text data.

This preprocessing enables the fast extraction of features from the text
data.

The preprocessing steps implemented in this module are:

- Tokenization:
    The text data is tokenized into individual words.
- Lemmatization:
    The words in the text data are converted to their base form.
- Syllabification:
    The words in the text data are split into syllables.
- POS Tagging:
    The words in the text data are tagged with their part-of-speech.
- Dependency Parsing:
    The words in the text data are parsed for dependencies.
- Named Entity Recognition:
    The named entities in the text data are identified.
"""
import polars as pl
import spacy
from spacy_syllables import SpacySyllables
import stanza

def preprocess_data(data: pl.DataFrame,
                    text_column: str = 'text',
                    backbone: str = 'spacy',
                    model: str = 'en_core_web_sm',
                    max_length: int = 1000000,
                    **kwargs: dict[str, str],
                    ) -> pl.DataFrame:
    """
    Preprocesses the text data using the specified NLP library.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        text_column: The name of the column containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.
        model (str): The name of the model used by the NLP library.
        max_length (int): The maximum number of characters to process.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the processed text data.
            The processed data is stored in a new column named 'nlp'.
    """
    if backbone == 'spacy':
        nlp = spacy.load(model)
        nlp.max_length = max_length
        nlp.add_pipe("syllables", after="tagger")
    elif backbone == 'stanza':
        nlp = stanza.Pipeline(model=model,
                              processors='tokenize,pos,lemma,depparse')
    
    # Process the text data to retrieve nlp objects
    processed = pl.Series("nlp", [nlp(text) for text in data[text_column]])

    # Insert the processed data into the DataFrame as the last column
    data = data.insert_column(len(data.columns), processed)
    
    return data

def get_lemmas(data: pl.DataFrame,
               backbone: str = 'spacy',
               **kwargs: dict[str, str],
               ) -> pl.DataFrame:
    """
    Gets the lemmas of the text data.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the lemmas of the text data.
            The lemmas are stored in a new column named 'lemmas'.
    """
    if backbone == 'spacy':
        lemmas = pl.Series("lemmas", [[token.lemma_ for token in doc]
                                      for doc in data['nlp']])
    elif backbone == 'stanza':
        lemmas = pl.Series("lemmas", [[word.lemma for sent in doc.sentences
                                      for word in sent.words]
                                      for doc in data['nlp']])
    data = data.insert_column(len(data.columns), lemmas)
    
    return data

def get_tokens(data: pl.DataFrame,
               backbone: str = 'spacy',
               **kwargs: dict[str, str],
               ) -> pl.DataFrame:
    """
    Gets the tokens of the text data.
    
    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        backbone (str): The NLP library used to process the text data.
                Either 'spacy' or 'stanza'.

    Returns:
        data (pl.DataFrame): 
            A Polars DataFrame containing the tokens of the text data.
            The tokens are stored in a new column named 'tokens'.
    """
    if backbone == 'spacy':
        tokens = pl.Series("tokens", [[token.text for token in doc]
                                      for doc in data['nlp']])
    elif backbone == 'stanza':
        tokens = pl.Series("tokens", [[word.text for sent in doc.sentences
                                      for word in sent.words]
                                      for doc in data['nlp']])
    data = data.insert_column(len(data.columns), tokens)
    
    return data




"""
This module contains schema definitions for the different external
resources. These schemas are used in Polars to enable type-safe and
fast data manipulation.
"""
import polars as pl


# ======================================== #
#        Sentiment/Emotion Analysis        #
# ======================================== #

# ---------------------------------------- #
#                   VAD                    #
# ---------------------------------------- #

VAD_SCHEMA_NRC = {
    "word": pl.String,
    "valence": pl.Float32,
    "arousal": pl.Float32,
    "dominance": pl.Float32,
}

VAD_SCHEMA_WARRINER = {
    '': pl.UInt32,
    'Word': pl.String,
    "V.Mean.Sum": pl.Float32,
    "V.SD.Sum": pl.Float32,
    "V.Rat.Sum": pl.Float32,
    "A.Mean.Sum": pl.Float32,
    "A.SD.Sum": pl.Float32,
    "A.Rat.Sum": pl.Float32,
    "D.Mean.Sum": pl.Float32,
    "D.SD.Sum": pl.Float32,
    "D.Rat.Sum": pl.Float32,
    "V.Mean.M": pl.Float32,
    "V.SD.M": pl.Float32,
    "V.Rat.M": pl.Float32,
    "V.Mean.F": pl.Float32,
    "V.SD.F": pl.Float32,
    "V.Rat.F": pl.Float32,
    "A.Mean.M": pl.Float32,
    "A.SD.M": pl.Float32,
    "A.Rat.M": pl.Float32,
    "A.Mean.F": pl.Float32,
    "A.SD.F": pl.Float32,
    "A.Rat.F": pl.Float32,
    "D.Mean.M": pl.Float32,
    "D.SD.M": pl.Float32,
    "D.Rat.M": pl.Float32,
    "D.Mean.F": pl.Float32,
    "D.SD.F": pl.Float32,
    "D.Rat.F": pl.Float32,
    "V.Mean.Y": pl.Float32,
    "V.SD.Y": pl.Float32,
    "V.Rat.Y": pl.Float32,
    "V.Mean.O": pl.Float32,
    "V.SD.O": pl.Float32,
    "V.Rat.O": pl.Float32,
    "A.Mean.Y": pl.Float32,
    "A.SD.Y": pl.Float32,
    "A.Rat.Y": pl.Float32,
    "A.Mean.O": pl.Float32,
    "A.SD.O": pl.Float32,
    "A.Rat.O": pl.Float32,
    "D.Mean.Y": pl.Float32,
    "D.SD.Y": pl.Float32,
    "D.Rat.Y": pl.Float32,
    "D.Mean.O": pl.Float32,
    "D.SD.O": pl.Float32,
    "D.Rat.O": pl.Float32,
    "V.Mean.L": pl.Float32,
    "V.SD.L": pl.Float32,
    "V.Rat.L": pl.Float32,
    "V.Mean.H": pl.Float32,
    "V.SD.H": pl.Float32,
    "V.Rat.H": pl.Float32,
    "A.Mean.L": pl.Float32,
    "A.SD.L": pl.Float32,
    "A.Rat.L": pl.Float32,
    "A.Mean.H": pl.Float32,
    "A.SD.H": pl.Float32,
    "A.Rat.H": pl.Float32,
    "D.Mean.L": pl.Float32,
    "D.SD.L": pl.Float32,
    "D.Rat.L": pl.Float32,
    "D.Mean.H": pl.Float32,
    "D.SD.H": pl.Float32,
    "D.Rat.H": pl.Float32,
}

# ---------------------------------------- #
#           Emotion Intensity              #
# ---------------------------------------- #

INTENSITY_SCHEMA = {
    "word": pl.String,
    "emotion": pl.String,
    "emotion_intensity": pl.Float32,
}

# ---------------------------------------- #
#               Sentiment                  #
# ---------------------------------------- #

SENTIWORDNET_SCHEMA = {
    "POS": pl.String,
    "ID": pl.UInt32,
    "PosScore": pl.Float32,
    "NegScore": pl.Float32,
    "SynsetTerms": pl.String,
    "Gloss": pl.String,
}

SENTIMENT_NRC_SCHEMA = {
    "word": pl.String,
    "emotion": pl.String,
    "label": pl.UInt8,
}

# ======================================== #
#        Psycholinguistic Features         #
# ======================================== #


import polars as pl

from .surface import (
    get_num_tokens,
    get_num_sentences,
    get_num_types
)

def get_feature_token_ratio(data: pl.DataFrame,
                            features: list[str],
                            backbone: str = 'spacy',
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Gets the ratio of given features to the total number of tokens.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        features (list[str]):
            A list of features to calculate the ratio for.
            Note that the features should be present in the data
            as column names.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the ratio of the given features
            to the total number of tokens. The ratios are stored in new
            columns with the feature names suffixed by '_token_ratio'. 
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone)

    data = data.with_columns(
        (pl.col(features) / pl.col("n_tokens")). \
            name.map(lambda x: f"{x}_token_ratio")
    )

    return data

def get_feature_type_ratio(data: pl.DataFrame,
                            features: list[str],
                            backbone: str = 'spacy',
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Gets the ratio of given features to the total number of types.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        features (list[str]):
            A list of features to calculate the ratio for.
            Note that the features should be present in the data
            as column names.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the ratio of the given features
            to the total number of types. The ratios are stored in new
            columns with the feature names suffixed by '_type_ratio'.
    """
    if 'n_types' not in data.columns:
        data = get_num_types(data, backbone)
    
    data = data.with_columns(
        (pl.col(features) / pl.col("n_types")). \
            name.map(lambda x: f"{x}_type_ratio")
    )

    return data

def get_feature_sentence_ratio(data: pl.DataFrame,
                               features: list[str],
                               backbone: str = 'spacy',
                               **kwargs: dict[str, str],
                               ) -> pl.DataFrame:
    """
    Gets the ratio of given features to the total number of sentences.

    Args:
        data (pl.DataFrame): A Polars DataFrame containing the text data.
        features (list[str]):
            A list of features to calculate the ratio for.
            Note that the features should be present in the data
            as column names.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the ratio of the given features
            to the total number of sentences. The ratios are stored in new
            columns with the feature names suffixed by '_sentence_ratio'. 
    """
    if 'n_sentences' not in data.columns:
        data = get_num_sentences(data, backbone)

    data = data.with_columns(
        (pl.col(features) / pl.col("n_sentences")). \
            name.map(lambda x: f"{x}_sentence_ratio")
    )

    return data 




"""
This module provides functions to extract semantic features from text.

The WordNet features are collected using the `wn` package. It uses Open
Multilingual WordNet to provide information about words in different
languages.

If you are using the resources for research, please cite the original
authors of Open Multilingual WordNet. 

The semantic features implemented in this module are:
- Number of hedge words
- Ratio of hedge words
- Average number of synsets
- Average number of synsets per POS tag
- Number of words with a low number of synsets
- Number of words with a high number of synsets
- Number of words with a low number of synsets per POS tag
- Number of words with a high number of synsets per POS tag
"""
import polars as pl
import wn

from .surface import (
    get_num_tokens,
)
from .util import (
    upos_to_wn,
)

# ------------------------------------------------------------------- #
#                            Hedges                                   #
# ------------------------------------------------------------------- #

def load_hedges(hedges_file: str) -> list[str]:
    """
    Loads the hedges from the given file.

    Args:
        hedges_file (str): Path to the file containing the hedges.

    Returns:
        hedges (list[str]):
            List of hedge words.
    """
    with open(hedges_file, 'r') as f:
        raw = f.readlines()
        hedges = []
        for line in raw:
            if not line.startswith("%") and line.strip() != "\n" and \
                line.strip() != "":
                hedges.append(line.strip())
    
    return hedges

def get_num_hedges(data: pl.DataFrame,
                   lexicon: list[str],
                   text_column: str = 'text',
                   **kwargs: dict[str, str],
                   ) -> pl.DataFrame:
    """
    Calculates the number of hedges in a text.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        lexicon (list[str]): List of hedges.
        text_column (str): Name of the column containing the text.

    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the number of hedges.
    """
    def n_matches(string: str, patterns: list[str]) -> int:
        """
        Helper function.
        Returns the number of matches of the patterns in the string.
        """
        total_matches = 0
        for pattern in patterns:
            total_matches += string.count(pattern)
        return total_matches
    
    data = data.with_columns(
        pl.col(text_column).map_elements(lambda x: 
                                         n_matches(x, lexicon),
            return_dtype=pl.UInt16
            ).alias("n_hedges"),
    )
    
    return data

def get_hedges_ratio(data: pl.DataFrame,
                     lexicon: list[str],
                     text_column: str = 'text',
                     backbone: str = 'spacy',
                     **kwargs: dict[str, str],
                     ) -> pl.DataFrame:
    """
    Calculates the ratio of hedges in a text.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        hedges (list[str]): List of hedges.
        text_column (str): Name of the column containing the text.
        backbone (str): NLP library used.

    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the hedges ratio.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    
    if 'n_hedges' not in data.columns:
        data = get_num_hedges(data, lexicon, text_column)
    
    data = data.with_columns(
        (pl.col("n_hedges") / pl.col("n_tokens")
         ).alias("hedges_ratio"),
    )
    
    return data

# --------------------------------------------------------------------- #
#                       Ambiguity, Polysemy                             #
# --------------------------------------------------------------------- #

def get_synsets(data: pl.DataFrame,
                backbone: str = 'spacy',
                language: str = 'en',
                pos_tags: list[str] = [
                    'NOUN', 'VERB', 'ADJ', 'ADV'
                    ],
                **kwargs: dict[str, str],
                ) -> pl.DataFrame:
    """
    Calculates the number of synsets in a text per token.

    WordNet synsets serve as a proxy for the ambiguity/polysemy of a word.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        backbone (str): NLP library used.
        language (str): Language of the text.
                Defaults to English ('en').
        pos_tags (list[str]): List of POS tags to consider.
                Defaults to lexical tokens.

    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the numbers of synsets per text.
            The columns are named 'synsets' and 'synsets_{pos}'.
    """
    if backbone == 'spacy':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x:
                                        [len(wn.synsets(
                                            token.text,
                                            lang=language,
                                            pos=upos_to_wn(token.pos_)))
                                         for token in x
                                         if token.pos_ in pos_tags],
                return_dtype=pl.List(pl.Int64)).alias("synsets")
        )
        for pos in pos_tags:
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x:
                                            [len(wn.synsets(
                                                token.text,
                                                lang=language,
                                                pos=upos_to_wn(token.pos_)
                                                ))
                                             for token in x
                                             if token.pos_ == pos],
                    return_dtype=pl.List(pl.Int64)). \
                        alias(f"synsets_{pos.lower()}")
            )
    elif backbone == 'stanza':
        data = data.with_columns(
            pl.col("nlp").map_elements(lambda x:
                                        [len(wn.synsets(
                                            word.text,
                                            lang=language,
                                            pos=upos_to_wn(word.upos)))
                                         for sent in x.sentences
                                         for word in sent.words
                                         if word.upos in pos_tags],
                return_dtype=pl.List(pl.Int64)).alias("synsets")
        )
        for pos in pos_tags:
            data = data.with_columns(
                pl.col("nlp").map_elements(lambda x:
                                            [len(wn.synsets(
                                                word.text,
                                                lang=language,
                                                pos=upos_to_wn(word.upos)))
                                             for sent in x.sentences
                                             for word in sent.words
                                             if word.upos == pos],
                    return_dtype=pl.List(pl.Int64)). \
                        alias(f"synsets_{pos.lower()}")
            )

    return data

def get_avg_num_synsets(data: pl.DataFrame,
                        backbone: str = 'spacy',
                        language: str = 'en',
                        pos_tags: list[str] = [
                            'NOUN', 'VERB', 'ADJ', 'ADV'
                            ],
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the average number of synsets in a text.

    WordNet synsets serve as a proxy for the ambiguity/polysemy of a word.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        backbone (str): NLP library used.
                'spacy' or 'stanza'.
        language (str): Language of the text.
                Defaults to English ('en').
        pos_tags (list[str]): List of POS tags to consider.
                Defaults to lexical tokens.

    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the average number of synsets.
            The column is named 'avg_n_synsets'.
    """
    if 'n_tokens' not in data.columns:
        data = get_num_tokens(data, backbone=backbone)
    if 'synsets' not in data.columns:
        data = get_synsets(data, backbone=backbone,
                           language=language,
                           pos_tags=pos_tags)
    
    data = data.with_columns(
        pl.col("synsets").list.mean().alias("avg_n_synsets")
    )
    
    
    return data

def get_avg_num_synsets_per_pos(data: pl.DataFrame,
                                backbone: str = 'spacy',
                                language: str = 'en',
                                pos_tags: list[str] = [
                                    'NOUN', 'VERB', 'ADJ', 'ADV'
                                    ],
                                nan_value: float = 0,
                                **kwargs: dict[str, str],
                                ) -> pl.DataFrame:
    """
    Calculates the average number of synsets per POS in a text.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        backbone (str): NLP library used.
                'spacy' or 'stanza'.
        language (str): Language of the text.
                Defaults to English ('en').
        pos_tags (list[str]): List of POS tags to consider.
                Defaults to lexical tokens.
        nan_value: Value to fill NaNs with.
                    Defaults to 0.

    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the average number of synsets per POS.
            The columns are named 'avg_n_synsets_{pos}'.
    """
    if "synsets" not in data.columns:  # ensures all synsets are calculated
        data = get_synsets(data, backbone=backbone,
                           language=language,
                           pos_tags=pos_tags)
    for pos_tag in pos_tags:
        if backbone == 'spacy':
            data = data.with_columns(
                pl.col("synsets_" + pos_tag.lower()).list.mean(). \
                fill_nan(nan_value).fill_null(nan_value). \
                    alias(f"avg_n_synsets_{pos_tag.lower()}")
            )
        elif backbone == 'stanza':
            data = data.with_columns(
                pl.col("synsets_" + pos_tag.lower()).list.mean(). \
                fill_nan(nan_value).fill_null(nan_value). \
                    alias(f"avg_n_synsets_{pos_tag.lower()}")
            )

    return data

def get_num_low_synsets(data: pl.DataFrame,
                        backbone: str = 'spacy',
                        language: str = 'en',
                        pos_tags: list[str] = [
                            'NOUN', 'VERB', 'ADJ', 'ADV'
                            ],
                        threshold: int = 2,
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the number of words with a low number of synsets in a text.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        backbone (str): NLP library used.
                'spacy' or 'stanza'.
        language (str): Language of the text.
                Defaults to English ('en').
        pos_tags (list[str]): List of POS tags to consider.
                Defaults to lexical tokens.
        threshold: Threshold for the number of synsets.
                    Defaults to 2.

    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the number of low synsets.
            The column is named 'n_low_synsets'.
    """
    if "synsets" not in data.columns:
        data = get_synsets(data, backbone=backbone,
                           language=language,
                           pos_tags=pos_tags)
    
    data = data.with_columns(
        pl.col("synsets").map_elements(lambda x:
                                       [
                                           1 for synset in x
                                           if synset <= threshold
                                       ],
        return_dtype=pl.List(pl.Int64)).list.len().alias("n_low_synsets")
    )


    return data

def get_num_high_synsets(data: pl.DataFrame,
                         backbone: str = 'spacy',
                         language: str = 'en',
                         pos_tags: list[str] = [
                             'NOUN', 'VERB', 'ADJ', 'ADV'
                             ],
                        threshold: int = 5,
                        **kwargs: dict[str, str],
                        ) -> pl.DataFrame:
    """
    Calculates the number of words with a high number of synsets in a text.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        backbone (str): NLP library used.
                'spacy' or 'stanza'.
        language (str): Language of the text.
                Defaults to English ('en').
        pos_tags (list[str]): List of POS tags to consider.
                Defaults to lexical tokens.
        threshold: Threshold for the number of synsets.
                    Defaults to 5.

    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the number of high synsets.
            The column is named 'n_high_synsets'.
    """
    if "synsets" not in data.columns:
        data = get_synsets(data, backbone=backbone,
                           language=language,
                           pos_tags=pos_tags)
    
    data = data.with_columns(
        pl.col("synsets").map_elements(lambda x:
                                       [
                                           1 for synset in x
                                           if synset >= threshold
                                       ],
        return_dtype=pl.List(pl.Int64)).list.len().alias("n_high_synsets")
    )

    return data

def get_low_synsets_per_pos(data: pl.DataFrame,
                            backbone: str = 'spacy',
                            language: str = 'en',
                            pos_tags: list[str] = [
                                'NOUN', 'VERB', 'ADJ', 'ADV'
                                ],
                            threshold: int = 2,
                            **kwargs: dict[str, str],
                            ) -> pl.DataFrame:
    """
    Calculates the number of words with a low number of synsets per POS in
    a text.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        backbone (str): NLP library used.
                'spacy' or 'stanza'.
        language (str): Language of the text.
                Defaults to English ('en').
        pos_tags (list[str]): List of POS tags to consider.
                Defaults to lexical tokens.
        threshold: Threshold for the number of synsets.
                 Defaults to 2.
    
    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the number of low synsets per POS.
            The columns are named 'n_low_synsets_{pos}'.
    """
    for pos_tag in pos_tags:
        if "synsets_" + pos_tag.lower() not in data.columns:
            data = get_synsets(data, backbone=backbone,
                               language=language,
                               pos_tags=[pos_tag])
        data = data.with_columns(
            pl.col("synsets_" + pos_tag.lower()).map_elements(
                lambda x: [1 for synset in x if synset <= threshold],
            return_dtype=pl.List(pl.Int64)).list.len(). \
                alias(f"n_low_synsets_{pos_tag.lower()}")
        )

    return data

def get_high_synsets_per_pos(data: pl.DataFrame,
                             backbone: str = 'spacy',
                             language: str = 'en',
                             pos_tags: list[str] = [
                                 'NOUN', 'VERB', 'ADJ', 'ADV'
                                 ],
                             threshold: int = 5,
                             **kwargs: dict[str, str],
                             ) -> pl.DataFrame:
    """
    Calculates the number of words with a high number of synsets per POS
    in a text.

    Args:
        data (pl.DataFrame): Polars DataFrame.
        backbone (str): NLP library used.
                'spacy' or 'stanza'.
        language (str): Language of the text.
                Defaults to English ('en').
        pos_tags (list[str]): List of POS tags to consider.
                Defaults to lexical tokens.
        threshold: Threshold for the number of synsets.
                 Defaults to 5.
    
    Returns:
        data (pl.DataFrame):
            Polars DataFrame with the number of high synsets per POS.
            The columns are named 'n_high_synsets_{pos}'.
    """
    for pos_tag in pos_tags:
        if "synsets_" + pos_tag.lower() not in data.columns:
            data = get_synsets(data, backbone=backbone,
                               language=language,
                               pos_tags=[pos_tag])
        data = data.with_columns(
            pl.col("synsets_" + pos_tag.lower()).map_elements(
                lambda x: [1 for synset in x if synset >= threshold],
            return_dtype=pl.List(pl.Int64)).list.len(). \
                alias(f"n_high_synsets_{pos_tag.lower()}")
        )

    return data




"""
This module contains functions to extract morphological features of text
data.

The morphological features implemented in this module are:

- Number of tokens with a specific morphological feature: The number of
    tokens that have a specific morphological feature, such as VerbForm,
    Number, etc.
"""
import polars as pl

from .configs.morphological_config import MORPH_CONFIG

def get_morph_feats(
        data: pl.DataFrame,
        backbone: str = 'spacy',
        morph_config: dict = MORPH_CONFIG,
        **kwargs: dict[str, str],
        ) -> pl.DataFrame:
    """
    Extracts morphological features from the text data.

    Args:
        data (pl.DataFrame): 
            A Polars DataFrame containing the text data.
        backbone (str): 
            The NLP library used to process the text data.
            Either 'spacy' or 'stanza'.
        morph_config (dict[str, str]):
            A dictionary containing the configuration for extracting
            morphological features. The keys are POS, and the values
            are dictionaries containing names of features as keys and
            their configurations as values. The configuration should
            follow the format as in the MORPH_CONFIG dictionary in
            elfen/configs/morphological_config.py.

    Returns:
        data (pl.DataFrame):
            A Polars DataFrame containing the morphological features of
            the text data.
    """
    if morph_config is None:
        print("No morphological features to extract. Returning the input"
              " data.")
        return data
    
    if backbone == 'spacy':
        for pos, feats in morph_config.items():
            for feat, values in feats.items():
                for val in values:
                    data = data.with_columns(
                        pl.col("nlp").map_elements(lambda x: len(
                            [token for token in x if token.pos_ == pos and
                            val in token.morph.get(feat)]),
                            return_dtype=pl.UInt16
                            ).alias(f"n_{pos}_{feat}_{val}"),
                    )
    elif backbone == 'stanza':
        for pos, feats in morph_config.items():
            for feat, values in feats.items():
                for val in values:
                    data = data.with_columns(
                        pl.col("nlp").map_elements(lambda x: len(
                            [1 for token in [token for sent in \
                                             x.sentences for token in \
                                             sent.words if \
                                             token.upos == pos and \
                                             token.feats]
                             if f"{feat}={val}" in token.feats]),
                            return_dtype=pl.UInt16
                            ).alias(f"n_{pos}_{feat}_{val}"),
                    )

    return data

